/*
    Copyright 2004,2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#ifdef WIN32
    #include <winsock.h>
    #include <direct.h>
    #include "winerr.h"

    #define    close    closesocket
    #define sleep   Sleep
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netdb.h>
#endif



#define VER    "0.1.5"
#define H416    0x36313420
#define H200    0x30303220
#define H206    0x36303220
#define PORT    80
#define PROXYP    8080
#define HTTP    "http://"
#define FTP    "ftp://"
#define BUFFSZ    4096
#define FILESZ    65535
#define ENDHEAD    "\r\n\r\n"
#define START    1
#define END    255
#define SEQSZ    30    /* size of possible maximum number (dir + file)*/
#define CLEN    "Content-Length: "
#define GET    "GET /%s HTTP/1.0\r\n" \
        "Host: %s\r\n" \
        "Range: bytes=%u-\r\n" \
        "Connection: close\r\n\r\n"
#define GETP    "GET http://%s:%d/%s HTTP/1.0\r\n" \
        "Host: %s\r\n" \
        "Range: bytes=%u-\r\n" \
        "Connection: close\r\n\r\n"



void get_url(u_char *buff);
void http2file(u_char *getstr);
void std_err(void);
u_int resolv(char *host);
void write_err(void);



u_int        currsz;
u_char        filename[BUFSIZ],
        *host,
        *uri;
u_short        port = PORT;
int        curr,
        showoff = 0;
struct    sockaddr_in    peer;



int main(int argc, char *argv[]) {
    u_char        getfile[BUFFSZ],
            dir[20],
            *dhost,
            *tmp1,
            *tmp,
            *proxystr = 0;
    int         err,
            fderr,
            start = START,
            end = END,
            dirstart = 0,
            dirend = 0,
            proxy = 0,
            tmplen;
    u_int        bytes = 0;
    u_short        dport;
    struct        stat    xstat;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif


    setbuf(stdout, NULL);

    printf("\n"
        "PDOWN "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n");

    if(argc < 2) {
        printf("\n"
            "Usage: %s [options] <\"URL()\">\n\n"
            "Options:\n"
            "-b BYTE           start the download from BYTE of the files\n"
            "-f FROM TO        file number start-end (default %d and %d)\n"
            "-d FROM TO        directory number start-end\n"
            "-p proxy[:port]   use proxy\n"
            "-o                during download show offset instead of percentage\n"
            "\n", argv[0], START, END);
        exit(1);
    }


    argc--;
    for(err = 1; err < argc; err++) {
        switch(argv[err][1]) {
            case 'b': bytes = atoi(argv[++err]); break;
            case 'f': {
                if(argv[++err]) start = atoi(argv[err]);
                if(argv[++err]) end = atoi(argv[err]);
                }; break;
            case 'd': {
                if(argv[++err]) dirstart = atoi(argv[err]);
                if(argv[++err]) dirend = atoi(argv[err]);
                }; break;
            case 'p': {
                proxystr = argv[++err];
                proxy = 1;
                }; break;
            case 'o': showoff = 1; break;
            default: {
                printf("\nError: \"%s\" wrong argument\n", argv[err]);
                exit(1);
                }; break;
        }
    }

    if(start > end) {
        printf("\nError: start file (%d) is major than end (%d)\n",
            start, end);
        exit(1);
    }

    if(dirstart > dirend) {
        printf("\nError: start directory (%d) is major than end (%d)\n",
            dirstart, dirend);
        exit(1);
    }

    get_url(argv[argc]);


    /* proxy */
    if(proxy) {
        dport = PROXYP;
        dhost = strchr(proxystr, ':');
        if(dhost) {
            *dhost = 0;
            dport = atoi(dhost + 1);
        }
        dhost = proxystr;
    } else {
        dhost = host;
        dport = port;
    }

    peer.sin_addr.s_addr = resolv(dhost);
    peer.sin_port        = htons(dport);
    peer.sin_family      = AF_INET;


    printf("\n"
        "Dest:  %s:%d\n"
        "Host:  %s:%d\n"
        "URI:   /%s\n"
        "Start: %d\n"
        "End:   %d\n"
        "\n",
        inet_ntoa(peer.sin_addr), dport,
        host, port,
        uri,
        start,
        end);


    tmplen = strlen(uri) + SEQSZ;
    tmp = malloc(tmplen);
    if(!tmp) std_err();

    for(; dirstart <= dirend; dirstart++) {

        if(dirstart) {
            /* directories management */
            snprintf(dir, 20, "%d", dirstart);
            printf("Directory: %s\n", dir);

#ifdef WIN32
    err = mkdir(dir);
#else
    err = mkdir(dir, 0);
#endif

            if(err < 0) perror("\nAlert");
            err = chdir(dir);
            if(err < 0) std_err();
        }

        for(curr = start; curr <= end; curr++) {

            /* if there is a dir we will calculate it in the string */
            if(dirstart && dirend) snprintf(tmp, tmplen, uri, dirstart, curr);
                else snprintf(tmp, tmplen, uri, curr);


            /* we search the filename in the URL */
            tmp1 = strrchr(tmp, '/');
            if(!tmp1) tmp1 = tmp;
                else tmp1++;
            snprintf(filename, BUFSIZ, "%d-%s", curr, tmp1);


            /* visualization of the filename and calculation of its current size */
            printf("%d) %s --> %s\n", curr, tmp1, filename);
            fderr = stat(filename, &xstat);
            if(fderr < 0) currsz = 0;
                else currsz = xstat.st_size;
            currsz += bytes;


            /* creation of the string and starting of the download */
            if(proxy) err = snprintf(getfile, BUFFSZ, GETP, host, port, tmp, host, currsz);
                else err = snprintf(getfile, BUFFSZ, GET, tmp, host, currsz);

            http2file(getfile);
        }

        if(dirstart) {
            /* go back in the parent directory */
            err = chdir("..");
            if(err < 0) std_err();
        }
    }

    printf("\n");
    return(0);
}



void http2file(u_char *getstr) {
    int        sock,
            err,
            length = 0,
            length2 = 0,
            realfile = 0,
            fderr;
    u_int        fsize = 0,
            perc = 0;
    fd_set        readset;
    u_char        *filebuff,
            *stri,
            *contlen;
    FILE        *fd = 0;


    sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(sock < 0) std_err();

    err = connect(sock, (struct sockaddr *)&peer, sizeof(peer));
    if(err < 0) std_err();

    err = strlen(getstr);
    send(sock, getstr, err, 0);


    filebuff = malloc(FILESZ);
    if(!filebuff) std_err();


    while(1) {
        FD_ZERO(&readset);
        FD_SET(sock, &readset);

        err = select(sock + 1, &readset, NULL, NULL, NULL);
        if(err < 0) std_err();

        if(FD_ISSET(sock, &readset)) {
            err = recv(sock, filebuff + length2,  FILESZ - length2, 0);
            if(err < 0) std_err();
            if(!err) break;
            length += err;

            if(realfile) {
                fderr = fwrite(filebuff, err, 1, fd);
                if(fderr != 1) write_err();
                if(fsize) {
                    if(!showoff) printf("%03u%%\r", (length + currsz) / perc);
                        else printf("%d (tot: %d)\r", length, length + currsz);
                } else printf(".");
            } else {

                filebuff[length] = 0;
                stri = strstr(filebuff, ENDHEAD);

                if(stri) {
                    stri += (sizeof(ENDHEAD) - 1);


                    /* check if file exist (HTTP/1.1 200 OK) */
                    err = *(int *)(filebuff + 8);
                    if(err == H416) return;
                    if((err != H200) && (err != H206)) {
                        printf("\nError: File not found on the server\n");
                        return;
                    }



                    /* filesize */
                    contlen = strstr(filebuff, CLEN);
                    if(contlen) {
                        fsize = atoi(contlen + sizeof(CLEN) - 1) + currsz;
                        perc = fsize / 100;

                        /* open file */
                        if((currsz == 0) || (currsz > fsize)) {
                            fd = fopen(filename, "wb");
                            if(!fd) std_err();
                        } else if(currsz < fsize) {
                            fd = fopen(filename, "r+b");
                            if(!fd) {
                                fd = fopen(filename, "wb");
                                if(!fd) std_err();
                            } else fseek(fd, 0, SEEK_END);
                        } else if(currsz == fsize) return;

                        printf("                           %u\r", fsize);
                    } else {
                        printf("                           no recovery\r");
                        fd = fopen(filename, "wb");
                        if(!fd) std_err();
                    }


                    /* we write on the file the latest bytes of the string */
                    length2 = 0;
                    length = filebuff + length - stri;
                    fderr = fwrite(stri, length, 1, fd);
                    if(fderr != 1) write_err();
                    realfile = 1;

                } else length2 = length;
            }
        }
    }
    if(fd) fclose(fd);
    close(sock);

    if(fsize) {
        if(length < fsize) curr--;
    }
    printf("\n");
}



void get_url(u_char *buff) {
    u_char        *tmp;

    if(!memcmp(buff, FTP, sizeof(FTP) - 1)) {
        printf("\nError: Unsupported protocol\n");
        exit(1);
    }
    host = strstr(buff, HTTP);
    if(!host) host = buff;
        else host += sizeof(HTTP) - 1;

    uri = strchr(host, '/');
    if(uri) {
        uri[0] = 0;
        uri++;
    } else uri = "";
    tmp = strchr(host, ':');
    if(tmp) {
        port = atoi(tmp + 1);
        tmp[0] = 0;
    }
}



u_int resolv(char *hostchar) {
    struct        hostent    *hp;
    u_int        host_ip;

    host_ip = inet_addr(hostchar);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(hostchar);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", hostchar);
            exit(1);
        } else host_ip = *(u_int *)(hp->h_addr);
    }

    return(host_ip);
}



void write_err(void) {
    printf("\nError: Write error\n");
    exit(1);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif



